@@ -100,6 +100,8 @@ struct _LmConnection {
 
 	LmConnectData *connect_data;
 
+	LmDisconnectReason disconnect_reason;
+
 	gint          ref_count;
 };
 
@@ -156,6 +158,8 @@ static LmHandlerResult connection_auth_reply (LmMessageHandler    *handler,
 					      LmMessage           *m,
 					      gpointer             user_data);
 
+static void     connection_stream_error      (LmConnection        *connection,
+					      LmMessage           *m);
 static void     connection_stream_received   (LmConnection        *connection, 
 					      LmMessage           *m);
 
@@ -171,7 +175,7 @@ static GSource * connection_create_source       (LmConnection *connection);
 static void      connection_signal_disconnect   (LmConnection *connection,
 						 LmDisconnectReason reason);
 
-static void     connection_do_connect           (LmConnectData *connect_data);
+static gboolean connection_do_connect           (LmConnectData *connect_data);
 static guint    connection_add_watch            (LmConnection  *connection,
 						 GIOChannel    *channel,
 						 GIOCondition   condition,
@@ -279,6 +283,10 @@ connection_handle_message (LmConnection *connection, LmMessage *m)
 		connection_stream_received (connection, m);
 		goto out;
 	}
+	else if (lm_message_get_type (m) == LM_MESSAGE_TYPE_STREAM_ERROR) {
+		connection_stream_error (connection, m);
+		goto out;
+	}
 	
 	id = lm_message_node_get_attribute (m->node, "id");
 	
@@ -469,7 +477,7 @@ _lm_connection_succeeded (LmConnectData *connect_data)
 	lm_message_unref (m);
 }
 
-void 
+gboolean 
 _lm_connection_failed_with_error (LmConnectData *connect_data, int error) 
 {
 	LmConnection *connection;
@@ -510,21 +518,19 @@ _lm_connection_failed_with_error (LmConnectData *connect_data, int error)
 							   cb->user_data);
 			_lm_utils_free_callback (cb);
 		}
-		
-		freeaddrinfo (connect_data->resolved_addrs);
-		connection->connect_data = NULL;
-		g_free (connect_data);
 	} else {
 		/* try to connect to the next host */
-		connection_do_connect (connect_data);
+		return connection_do_connect (connect_data);
 	}
+
+	return FALSE;
 }
 
-void 
+gboolean
 _lm_connection_failed (LmConnectData *connect_data)
 {
-	_lm_connection_failed_with_error (connect_data, 
-					  _lm_sock_get_last_error());
+	return _lm_connection_failed_with_error (connect_data, 
+						 _lm_sock_get_last_error());
 }
 
 static gboolean 
@@ -589,10 +595,13 @@ connection_connect_cb (GIOChannel   *source,
 		/* for blocking sockets, G_IO_OUT means we are connected */
 		g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
 		       "Connection success.\n");
-		
+
 		_lm_connection_succeeded (connect_data);
 	}
 
+	/* set the default disconnect reason to unknown */
+	connection->disconnect_reason = LM_DISCONNECT_REASON_UNKNOWN;
+
  	return TRUE; 
 }
 
@@ -617,7 +626,7 @@ connection_condition_to_str (GIOCondition condition)
 	return buf;
 }
 
-static void
+static gboolean
 connection_do_connect (LmConnectData *connect_data) 
 {
 	LmConnection    *connection;
@@ -646,8 +655,7 @@ connection_do_connect (LmConnectData *connect_data)
 			   NI_NUMERICHOST | NI_NUMERICSERV);
 	
 	if (res < 0) {
-		_lm_connection_failed (connect_data);
-		return;
+		return _lm_connection_failed (connect_data);
 	}
 
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
@@ -662,9 +670,7 @@ connection_do_connect (LmConnectData *connect_data)
 		       "Failed making socket, error:%d...\n",
 		       _lm_sock_get_last_error ());
 
-		_lm_connection_failed (connect_data);
-
-		return;
+		return _lm_connection_failed (connect_data);
 	}
 
 	/* Even though it says _unix_new(), it is supported by glib on
@@ -704,11 +710,11 @@ connection_do_connect (LmConnectData *connect_data)
 		err = _lm_sock_get_last_error ();
 		if (!_lm_sock_is_blocking_error (err)) {
 			_lm_sock_close (connect_data->fd);
-			_lm_connection_failed_with_error (connect_data, err);
-
-			return;
+			return _lm_connection_failed_with_error (connect_data, err);
 		}
 	}
+
+	return TRUE;
 }
 
 static guint
@@ -948,7 +954,14 @@ connection_do_open (LmConnection *connection, GError **error)
 
 	connection->connect_data = data;
 
-	connection_do_connect (data);
+	if (!connection_do_connect (data)) {
+		g_set_error (error,
+			LM_ERROR,                 
+			LM_ERROR_CONNECTION_FAILED,
+			"unable to connect");
+		return FALSE;
+	}
+
 	return TRUE;
 }
 					
@@ -1106,16 +1119,23 @@ connection_read_incoming (LmConnection *connection,
 		switch (status) {
 		case G_IO_STATUS_EOF:
 			reason = LM_DISCONNECT_REASON_HUP;
+			lm_verbose ("Disconnect reason: %d\n", reason);
 			break;
 		case G_IO_STATUS_AGAIN:
 			/* No data readable but we didn't hangup */
 			return FALSE;
 			break;
 		case G_IO_STATUS_ERROR:
-			reason = LM_DISCONNECT_REASON_ERROR;
-			break;
+			/* If no reason set, we set it to ERROR here,
+			 * otherwise it might already have been set by
+			 * the stream error handler.
+			 */
+			if (connection->disconnect_reason == LM_DISCONNECT_REASON_UNKNOWN) {
+				connection->disconnect_reason = LM_DISCONNECT_REASON_ERROR;
+			}
 		default:
-			reason = LM_DISCONNECT_REASON_UNKNOWN;
+			reason = connection->disconnect_reason;
+			lm_verbose ("Disconnect reason: %d\n", reason);
 		}
 
 		connection_do_close (connection);
@@ -1438,6 +1458,35 @@ connection_auth_reply (LmMessageHandler *handler,
 	return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 }
 
+static void
+connection_stream_error (LmConnection *connection, LmMessage *m)
+{
+	LmMessageNode *node;
+
+	g_return_if_fail (connection != NULL);
+	g_return_if_fail (m != NULL);
+
+	node = m->node;
+
+	/* Resource conflict */
+	node = lm_message_node_get_child (node, "conflict");
+	if (node) {
+		lm_verbose ("Stream error: Conflict (resource connected elsewhere)\n");
+		connection->disconnect_reason = LM_DISCONNECT_REASON_RESOURCE_CONFLICT;
+		return;
+	}
+
+	/* XML is crack */
+	node = lm_message_node_get_child (node, "xml-not-well-formed");
+	if (node) {
+		lm_verbose ("Stream error: XML not well formed\n");
+		connection->disconnect_reason = LM_DISCONNECT_REASON_INVALID_XML;
+		return;
+	}
+
+	lm_verbose ("Stream error: Unrecognised error\n");
+	connection->disconnect_reason = LM_DISCONNECT_REASON_ERROR;
+}
 
 static void
 connection_stream_received (LmConnection *connection, LmMessage *m)
@@ -1469,7 +1518,7 @@ connection_stream_received (LmConnection *connection, LmMessage *m)
 			        			   cb->user_data);
 
 		}
-		_lm_utils_free_callback (connection->open_cb);
+		_lm_utils_free_callback (cb);
 	}
 }
 
@@ -1583,6 +1632,7 @@ lm_connection_new (const gchar *server)
 							 g_str_equal,
 							 g_free, 
 							 (GDestroyNotify) lm_message_handler_unref);
+	connection->disconnect_reason = LM_DISCONNECT_REASON_UNKNOWN;
 	connection->ref_count         = 1;
 	
 	for (i = 0; i < LM_MESSAGE_TYPE_UNKNOWN; ++i) {
